import { Request, Response } from "express";
import { handleErrors, parseRequest } from "../utils";
import { productRepository } from "./repository";
import {
  createProductRequestSchema,
  deleteProductRequestSchema,
  getProductRequestSchema,
  getProductsRequestSchema,
} from "./validationSchemas";

/*
Note that we use the same types for the database entities and the objects that are sent via the API.
For larger projects, it is preferable to use separate types for database entities and
data transfer objects (DTOs) because they may not always match.
*/

const createProduct = async (req: Request, res: Response) => {
  const request = await parseRequest(createProductRequestSchema, req, res);
  if (request === null) return;

  const product = request.body;

  const newProduct = await productRepository.create(product);
  if (newProduct.isErr) {
    handleErrors(newProduct.error, res);
    return;
  }

  res.status(201).send(newProduct.value);
};

const getProducts = async (req: Request, res: Response) => {
  const request = await parseRequest(getProductsRequestSchema, req, res);
  if (request === null) return;

  const filter = request.query;

  const products = await productRepository.getAll(filter);
  if (products.isErr) {
    handleErrors(products.error, res);
    return;
  }

  res.send(products.value);
};

const getProduct = async (req: Request, res: Response) => {
  const request = await parseRequest(getProductRequestSchema, req, res);
  if (request === null) return;

  const id = request.params.id;

  const product = await productRepository.getById(id);
  if (product.isErr) {
    handleErrors(product.error, res);
    return;
  }

  res.send(product.value);
};

const deleteProduct = async (req: Request, res: Response) => {
  const request = await parseRequest(deleteProductRequestSchema, req, res);
  if (request === null) return;

  const id = request.params.id;

  const result = await productRepository.delete(id);
  if (result.isErr) {
    handleErrors(result.error, res);
    return;
  }

  res.status(204).send("Product deleted");
};

export const productsController = {
  createProduct,
  deleteProduct,
  getProduct,
  getProducts,
};
