import { Result } from "@badrap/result";
import { redisClient } from "../redisClient";
import { InternalError, NotFoundError } from "../utils";
import { CreateProduct, GetProductsFilter, Product } from "./types";

const PRODUCT_KEY_PREFIX = "product";
const PRODUCT_IDS_SET_KEY = "productIds";
const PRODUCT_ID_COUNTER_KEY = "product:nextId";

export const productRepository = {
  async create(product: CreateProduct): Promise<Result<Product>> {
    try {
      const newProductId = await redisClient.incr(PRODUCT_ID_COUNTER_KEY);
      const newProduct: Product = { id: newProductId, ...product };

      await redisClient.set(
        `${PRODUCT_KEY_PREFIX}:${newProductId}`,
        JSON.stringify(newProduct)
      );

      await redisClient.sAdd(PRODUCT_IDS_SET_KEY, newProductId.toString());

      return Result.ok(newProduct);
    } catch (error) {
      return Result.err(new InternalError());
    }
  },

  async getAll(filter?: GetProductsFilter): Promise<Result<Product[]>> {
    try {
      const productIds = await redisClient.sMembers(PRODUCT_IDS_SET_KEY);

      const products: Product[] = [];

      for (const productId of productIds) {
        const productJson = await redisClient.get(
          `${PRODUCT_KEY_PREFIX}:${productId}`
        );
        if (!productJson) continue;

        const product = JSON.parse(productJson) as Product;

        // apply filter
        if (
          (filter?.name === undefined || product.name.includes(filter.name)) &&
          (filter?.description === undefined ||
            product.description?.includes(filter.description)) &&
          (filter?.minPrice === undefined ||
            product.price >= filter.minPrice) &&
          (filter?.maxPrice === undefined || product.price <= filter.maxPrice)
        ) {
          products.push(product);
        }
      }

      return Result.ok(products);
    } catch (error) {
      return Result.err(new InternalError());
    }
  },

  async getById(id: number): Promise<Result<Product>> {
    try {
      const productJson = await redisClient.get(`${PRODUCT_KEY_PREFIX}:${id}`);
      if (!productJson) {
        return Result.err(new NotFoundError(`Product with id ${id} not found`));
      }

      return Result.ok(JSON.parse(productJson) as Product);
    } catch (error) {
      return Result.err(new InternalError());
    }
  },

  async delete(id: number): Promise<Result<null>> {
    try {
      const removeKeyResult = await redisClient.sRem(
        PRODUCT_IDS_SET_KEY,
        id.toString()
      );
      if (!removeKeyResult) {
        return Result.err(new NotFoundError(`Product with id ${id} not found`));
      }

      const deleteValueResult = await redisClient.del(
        `${PRODUCT_KEY_PREFIX}:${id}`
      );
      if (!deleteValueResult) {
        return Result.err(new InternalError());
      }

      return Result.ok(null);
    } catch (error) {
      return Result.err(new InternalError());
    }
  },
};
